"""
A factory class for creating instances of AdditiveEvent and its descendants.
"""
from dicetables.dicerecord import DiceRecord
from dicetables.eventsbases.eventerrors import InvalidEventsError, DiceRecordError
from dicetables.factory.errorhandler import EventsFactoryErrorHandler
from dicetables.factory.factorytools import StaticDict, Getter
from dicetables.factory.warninghandler import EventsFactoryWarningHandler


class LoaderError(AttributeError):
    pass


class Loader(object):
    def __init__(self, factory):
        self.factory = factory

    def load(self, new_class):
        if 'factory_keys' in new_class.__dict__:
            factory_keys = new_class.factory_keys
        else:
            raise LoaderError

        new_keys = ()
        if 'new_keys' in new_class.__dict__:
            new_keys = new_class.new_keys

        for key in new_keys:
            self.factory.add_getter(*key)
        self.factory.add_class(new_class, factory_keys)


class EventsFactory(object):
    __default_getters = {'get_dict': Getter('get_dict', {0: 1}),
                         'dice_data': Getter('dice_data', DiceRecord.new()),
                         'calc_includes_zeroes': Getter('calc_includes_zeroes', True, is_property=True)}

    __default_class_args = {'AdditiveEvents': ('get_dict',),
                            'DiceTable': ('get_dict', 'dice_data'),
                            'DetailedDiceTable': ('get_dict', 'dice_data', 'calc_includes_zeroes')}

    _getters = StaticDict(__default_getters)
    _class_args = StaticDict(__default_class_args)

    @classmethod
    def reset(cls):
        cls._getters = StaticDict(cls.__default_getters)
        cls._class_args = StaticDict(cls.__default_class_args)

    @classmethod
    def has_class(cls, events_class):
        return events_class.__name__ in cls._class_args.keys()

    @classmethod
    def has_getter(cls, getter_key):
        return getter_key in cls._getters.keys()

    @classmethod
    def get_class_params(cls, events_class):
        return cls._class_args.get(events_class.__name__)

    @classmethod
    def get_getter_string(cls, getter_key):
        return cls._getters.get(getter_key).__str__()

    @classmethod
    def get_keys(cls):
        return sorted(cls._class_args.keys()), sorted(cls._getters.keys())

    @classmethod
    def add_class(cls, events_class, getter_key_words):
        cls._check_against_factory_classes(events_class, getter_key_words)
        cls._check_for_missing_getters(events_class, getter_key_words)
        cls._class_args = cls._class_args.set(events_class.__name__, getter_key_words)

    @classmethod
    def _check_against_factory_classes(cls, events_class, class_args_tuple):
        if cls.has_class(events_class) and cls.get_class_params(events_class) != class_args_tuple:
            EventsFactoryErrorHandler(cls).raise_error('CLASS OVERWRITE', events_class, class_args_tuple)

    @classmethod
    def _check_for_missing_getters(cls, events_class, class_args_tuple):
        for getter_key in class_args_tuple:
            if not cls.has_getter(getter_key):
                EventsFactoryErrorHandler(cls).raise_error('MISSING GETTER', events_class, getter_key)

    @classmethod
    def add_getter(cls, getter_name, empty_value, type_str='method'):
        is_property = False
        if type_str == 'property':
            is_property = True
        new_getter = Getter(getter_name, empty_value, is_property)
        cls._check_against_factory_getters(getter_name, new_getter)
        cls._getters = cls._getters.set(getter_name, new_getter)

    @classmethod
    def _check_against_factory_getters(cls, getter_key, new_getter):
        if cls.has_getter(getter_key) and cls._getters.get(getter_key) != new_getter:
            EventsFactoryErrorHandler(cls).raise_error('GETTER OVERWRITE', getter_key, new_getter)

    @classmethod
    def check(cls, events_class):
        if not cls.has_class(events_class):
            try:
                Loader(cls).load(events_class)
            except LoaderError:
                EventsFactoryWarningHandler(cls).raise_warning('CHECK', events_class)

    @classmethod
    def from_dictionary(cls, events, dictionary):
        passed_in_values = {'get_dict': dictionary}
        return cls._construct_from(events, passed_in_values)

    @classmethod
    def from_dictionary_and_dice(cls, events, dictionary, dice):
        passed_in_values = {'get_dict': dictionary, 'dice_data': dice}
        return cls._construct_from(events, passed_in_values)

    @classmethod
    def from_params(cls, events, name_value_param_dict):
        return cls._construct_from(events, name_value_param_dict)

    @classmethod
    def _construct_from(cls, events, passed_in_values):
        constructor_class = events.__class__
        factory_class = cls._get_nearest_factory_class(constructor_class)
        args = cls._get_args(events, factory_class, passed_in_values)
        return cls._construct(constructor_class, factory_class, args)

    @classmethod
    def new(cls, events_class):
        factory_class = cls._get_nearest_factory_class(events_class)
        args = cls._get_default_args(factory_class)
        return cls._construct(events_class, factory_class, args)

    @classmethod
    def _get_nearest_factory_class(cls, events_class):
        if not cls.has_class(events_class):
            try:
                Loader(cls).load(events_class)
            except LoaderError:
                return cls._walk_the_mro(events_class)
        return events_class

    @classmethod
    def _walk_the_mro(cls, failed_class):
        for parent_class in failed_class.mro():
            if cls.has_class(parent_class):
                EventsFactoryWarningHandler(cls).raise_warning('CONSTRUCT', failed_class, parent_class)
                return parent_class
        EventsFactoryErrorHandler(cls).raise_error('WTF', failed_class)

    @classmethod
    def _get_default_args(cls, in_factory):
        args = []
        for getter_key in cls.get_class_params(in_factory):
            getter = cls._getters.get(getter_key)
            args.append(getter.get_default())
        return args

    @classmethod
    def _get_args(cls, events, factory_class, passed_in_values):
        new_args = []
        for getter_key in cls.get_class_params(factory_class):
            if getter_key in passed_in_values.keys():
                arg_value = passed_in_values[getter_key]
            else:
                getter = cls._getters.get(getter_key)
                arg_value = getter.get_from(events)
            new_args.append(arg_value)
        return new_args

    @classmethod
    def _construct(cls, original_class, factory_class, args):
        try:
            return original_class(*args)
        except (TypeError, AttributeError, DiceRecordError, InvalidEventsError):
            if original_class == factory_class:
                EventsFactoryErrorHandler(cls).raise_error('SIGNATURES DIFFERENT', original_class)
            return factory_class(*args)



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import

import unittest

from dicetables.eventsbases.protodie import ProtoDie


class DummyDie(ProtoDie):
    def __init__(self, size, weight, dictionary, repr_str):
        self.size = size
        self.weight = weight
        self.dictionary = dictionary
        self.repr_str = repr_str
        super(DummyDie, self).__init__()

    def get_dict(self):
        return self.dictionary

    def get_size(self):
        return self.size

    def get_weight(self):
        return self.weight

    def __repr__(self):
        return self.repr_str

    def __str__(self):
        return 'hi'

    def multiply_str(self, number):
        return str(number)

    def weight_info(self):
        return "OyVei! you're too thin. Es, bubelah."


class TestProtoDie(unittest.TestCase):

    def test_ProtoDie_not_implemented_errors(self):
        self.assertRaises(NotImplementedError, ProtoDie)

        class TestDie(ProtoDie):
            def get_dict(self):
                return {1: 1}

        test = TestDie()
        self.assertRaises(NotImplementedError, test.get_size)
        self.assertRaises(NotImplementedError, test.get_weight)
        self.assertRaises(NotImplementedError, test.weight_info)
        self.assertRaises(NotImplementedError, test.multiply_str, 3)
        self.assertRaises(NotImplementedError, test.__str__)
        self.assertRaises(NotImplementedError, test.__repr__)

    # rich comparison testing
    def test_ProtoDie_equality_true(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(2, 3, {2: 2}, 'b')
        self.assertTrue(first == second)

    def test_ProtoDie_equality_false_if_different_types_eval_the_same(self):
        class NewDummyDie(DummyDie):
            def __init__(self, size, weight, dictionary, repr_str):
                super(NewDummyDie, self).__init__(size, weight, dictionary, repr_str)

        first = DummyDie(2, 3, {2: 2}, 'b')
        second = NewDummyDie(2, 3, {2: 2}, 'b')
        self.assertFalse(first == second)

    def test_ProtoDie_equality_false_by_size(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(100, 3, {2: 2}, 'b')
        self.assertFalse(first == second)

    def test_ProtoDie_equality_false_by_weight(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(2, 100, {2: 2}, 'b')
        self.assertFalse(first == second)

    def test_ProtoDie_equality_false_by_get_dict(self):
        first = DummyDie(2, 3, {200: 200}, 'b')
        second = DummyDie(2, 3, {2: 2}, 'b')
        self.assertFalse(first == second)

    def test_ProtoDie_equality_false_by_get_repr(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(2, 3, {2: 2}, 'different repr')
        self.assertFalse(first == second)

    def test_ProtoDie_equality_false_by_not_a_proto_die(self):
        self.assertFalse(DummyDie(2, 3, {2: 2}, 'b') == 2)

    def test_lt_false_if_equal(self):
        first = DummyDie(2, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertFalse(first < second)
        self.assertFalse(second < first)

    def test_ProtoDie_lt_by_size(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(20, 3, {2: 2}, 'b')
        self.assertTrue(first < second)

    def test_ProtoDie_lt_by_weight(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(2, 30, {2: 2}, 'b')
        self.assertTrue(first < second)

    def test_ProtoDie_lt_by_sorted_get_dict(self):
        first = DummyDie(2, 3, {2: 200, 20: 200}, 'b')
        second = DummyDie(2, 3, {20: 2}, 'b')
        third = DummyDie(2, 3, {20: 20}, 'b')
        self.assertTrue(first < second)
        self.assertTrue(first < third)
        self.assertTrue(second < third)

    def test_ProtoDie_lt_by_repr(self):
        first = DummyDie(2, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'b')
        self.assertTrue(first < second)

    def test_ProtoDie_lt_by_size_is_most_important(self):
        first = DummyDie(2, 30, {20: 20}, 'b')
        second = DummyDie(20, 3, {2: 2}, 'a')
        self.assertTrue(first < second)

    def test_ProtoDie_lt_by_weight_is_second_most_important(self):
        first = DummyDie(2, 3, {20: 20}, 'b')
        second = DummyDie(2, 30, {2: 2}, 'a')
        self.assertTrue(first < second)

    def test_ProtoDie_lt_by_sorted_get_dict_is_third_most_important_and_so_repr_is_last(self):
        first = DummyDie(2, 3, {2: 2}, 'b')
        second = DummyDie(2, 3, {20: 20}, 'a')
        self.assertTrue(first < second)

    def test_ProtoDie_not_equal_true(self):
        first = DummyDie(20, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertTrue(first != second)
        self.assertFalse(first == second)

    def test_ProtoDie_not_equal_false(self):
        first = DummyDie(2, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertTrue(first == second)
        self.assertFalse(first != second)

    def test_ProtoDie_gt_true(self):
        first = DummyDie(20, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertTrue(first > second)

    def test_ProtoDie_gt_false(self):
        first = DummyDie(20, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertFalse(second > first)
        self.assertFalse(first > first)

    def test_ProtoDie_le_true(self):
        first = DummyDie(2, 3, {2: 2}, 'a')
        second = DummyDie(20, 3, {2: 2}, 'a')
        self.assertTrue(first <= second)
        self.assertTrue(first <= first)

    def test_ProtoDie_le_false(self):
        first = DummyDie(2, 3, {2: 2}, 'a')
        second = DummyDie(20, 3, {2: 2}, 'a')
        self.assertFalse(second <= first)

    def test_ProtoDie_ge_true(self):
        first = DummyDie(20, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertTrue(first >= second)
        self.assertTrue(first >= first)

    def test_ProtoDie_ge_false(self):
        first = DummyDie(20, 3, {2: 2}, 'a')
        second = DummyDie(2, 3, {2: 2}, 'a')
        self.assertFalse(second >= first)

    def test_ProtoDie_hash(self):
        hash_str = 'hash of REPR, 2, 3, {4: 5}'
        self.assertEqual(hash(DummyDie(2, 3, {4: 5}, 'REPR')), hash(hash_str))


if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import

import unittest
import warnings

from dicetables.additiveevents import AdditiveEvents
from dicetables.dicetable import DiceTable
from dicetables.factory.eventsfactory import EventsFactory
from dicetables.factory.warninghandler import EventsFactoryWarning, EventsFactoryWarningHandler


class TestEventsFactoryWarningHandler(unittest.TestCase):
    def setUp(self):
        EventsFactory.reset()

    def assert_warning(self, warning, msg, func, *args):
        with warnings.catch_warnings(record=True) as cm:
            warnings.simplefilter("always")
            func(*args)
        if len(cm) != 1:
            self.fail('Warning not raised or too many warnings')
        else:
            self.assertEqual(msg, str(cm[0].message))
            self.assertEqual(warning, cm[0].category)

    def test_assert_warning(self):
        def func(number):
            warnings.warn('msg', EventsFactoryWarning, stacklevel=2)
            return number

        self.assert_warning(EventsFactoryWarning, 'msg', func, 5)

    def test_assert_warning_with_no_warning(self):
        def func(number):
            return number

        with self.assertRaises(AssertionError) as cm:
            self.assert_warning(EventsFactoryWarning, 'hi', func, 5)
        self.assertEqual(cm.exception.args[0], 'Warning not raised or too many warnings')

    def test_assert_warning_with_too_many_warnings(self):
        def func(number):
            warnings.warn('msg', EventsFactoryWarning, stacklevel=2)
            warnings.warn('msg', Warning, stacklevel=2)
            return number

        with self.assertRaises(AssertionError) as cm:
            self.assert_warning(EventsFactoryWarning, 'hi', func, 5)
        self.assertEqual(cm.exception.args[0], 'Warning not raised or too many warnings')

    def test_create_message_header(self):
        expected = (
                '\n' +
                'factory: <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Warning code: CHECK\n' +
                'Failed to find/add the following class to the EventsFactory - \n' +
                'class: <class \'dicetables.factory.warninghandler.EventsFactoryWarningHandler\'>\n'
        )
        self.assertEqual(create_message_header('CHECK', EventsFactory, EventsFactoryWarningHandler), expected)

    def test_create_error_code_body_CONSTRUCT(self):
        expected = (
                '\n' +
                'Class found in factory: <class \'dicetables.additiveevents.AdditiveEvents\'>\n' +
                'attempted object construction using its signature. tried to return instance of original class.\n' +
                'If that had failed, returned instance of the class found in EventsFactory.\n' +
                '\n'
        )
        self.assertEqual(create_error_code_body((EventsFactoryWarningHandler, AdditiveEvents), 'CONSTRUCT'),
                         expected)

    def test_create_error_code_body_CHECK(self):
        expected = '\nWarning raised while performing check at instantiation\n\n'
        self.assertEqual(create_error_code_body((EventsFactoryWarningHandler, AdditiveEvents), 'CHECK'),
                         expected)

    def test_create_instructions(self):
        self.maxDiff = None
        expected = (
                'SOLUTION:\n' +
                '  class variable: factory_keys = (getter method/property names)\n'
                '  current factory keys are: [\'calc_includes_zeroes\', \'dice_data\', \'get_dict\']\n' +
                '  class variable: new_keys = [(info for each key not already in factory)]\n' +
                '  Each tuple in "new_keys" is (getter_name, default_value, "property"/"method")\n' +
                'ex:\n' +
                '  NewClass(Something):\n' +
                '      factory_keys = ("dice_data", "get_dict", "get_thingy", "label")\n' +
                '      new_keys = [("get_thingy", 0, "method"),\n' +
                '                  ("label", "", "property")]\n' +
                '\n' +
                '      def __init__(self, events_dict, dice_list, new_thingy, label):\n' +
                '          ....\n'
        )
        self.assertEqual(create_instructions(EventsFactory), expected)

    def test_create_warning_message(self):
        EventsFactory.reset()
        factory = EventsFactory
        code = 'CHECK'
        failed_class = EventsFactoryWarningHandler
        found_class = AdditiveEvents

        start = create_message_header(code, factory, failed_class)
        middle = create_error_code_body((failed_class, found_class), code)
        end = create_instructions(factory)
        self.assertEqual(create_warning_message(factory, code, failed_class, found_class), start + middle + end)

    def test_FactoryWarningHandler_warning_by_CHECK(self):
        msg = create_warning_message(EventsFactory, 'CHECK', DiceTable, AdditiveEvents)
        self.assert_warning(EventsFactoryWarning, msg, EventsFactoryWarningHandler(EventsFactory).raise_warning,
                            'CHECK', DiceTable, AdditiveEvents)

    def test_FactoryWarningHandler_warning_by_CONSTRUCT(self):
        msg = create_warning_message(EventsFactory, 'CONSTRUCT', DiceTable, AdditiveEvents)
        self.assert_warning(EventsFactoryWarning, msg, EventsFactoryWarningHandler(EventsFactory).raise_warning,
                            'CONSTRUCT', DiceTable, AdditiveEvents)


def create_warning_message(factory, code, *classes):
    failed_class = classes[0]

    intro = create_message_header(code, factory, failed_class)
    middle = create_error_code_body(classes, code)

    instructions = create_instructions(factory)
    return intro + middle + instructions


def create_message_header(code, factory, failed_class):
    intro = (
            '\nfactory: {}\n'.format(factory) +
            'Warning code: {}\n'.format(code) +
            'Failed to find/add the following class to the EventsFactory - \n' +
            'class: {}\n'.format(failed_class)
    )
    return intro


def create_error_code_body(classes, code):
    if code == 'CONSTRUCT':
        in_factory = classes[1]
        middle = (
                '\nClass found in factory: {}\n'.format(in_factory) +
                'attempted object construction using its signature. tried to return instance of original class.\n' +
                'If that had failed, returned instance of the class found in EventsFactory.\n\n'
        )
    else:
        middle = '\nWarning raised while performing check at instantiation\n\n'
    return middle


def create_instructions(factory):
    instructions = (
            'SOLUTION:\n' +
            '  class variable: factory_keys = (getter method/property names)\n'
            '  current factory keys are: {}\n'.format(factory.get_keys()[1]) +
            '  class variable: new_keys = [(info for each key not already in factory)]\n' +
            '  Each tuple in "new_keys" is (getter_name, default_value, "property"/"method")\n'
            'ex:\n' +
            '  NewClass(Something):\n' +
            '      factory_keys = ("dice_data", "get_dict", "get_thingy", "label")\n' +
            '      new_keys = [("get_thingy", 0, "method"),\n' +
            '                  ("label", "", "property")]\n\n' +
            '      def __init__(self, events_dict, dice_list, new_thingy, label):\n' +
            '          ....\n'
    )
    return instructions


if __name__ == '__main__':
    unittest.main()



from typing import Dict

from dicetables.dieevents import ProtoDie
from dicetables.dicepool import DicePool


class DicePoolCollection(ProtoDie):
    """
    The abstract class for all DicePoolCollection objects. A DicePoolCollection creates a new die by selecting from a
    DicePool. Select determines how many rolls are selected from the pool of total rolls.
    Different implementations determine which particular rolls to select.
    """

    def __init__(self, pool: DicePool, select: int):
        """

        :param pool: the dice pool to select from
        :param select: int <= `pool.size`. How many dice to select from the pool.
        """
        self._dice_pool = pool
        if select > self._dice_pool.size:
            raise ValueError('you cannot select more dice than the pool_size.')
        self._select = select
        self._dict = self._generate_dict()
        super(DicePoolCollection, self).__init__()

    def _generate_dict(self):
        raise NotImplementedError

    def get_pool(self) -> DicePool:
        return self._dice_pool

    def get_select(self) -> int:
        return self._select

    def get_dict(self) -> Dict[int, int]:
        return self._dict.copy()

    def get_size(self) -> int:
        return self._dice_pool.die.get_size() * self._select

    def get_weight(self) -> int:
        return self._dice_pool.die.get_weight() * self._select

    def weight_info(self):
        input_die = self._dice_pool.die
        original_weight_info = input_die.weight_info().replace(str(input_die), 'input_die info:')
        return '{}\n{}'.format(self, original_weight_info)

    def multiply_str(self, number):
        return '{}({})'.format(number, self)

    def __str__(self):
        descriptor = self.__class__.__name__.replace('OfDicePool', '')
        return '{} {} of {}'.format(descriptor, self._select, self._dice_pool.die.multiply_str(self._dice_pool.size))

    def __repr__(self):
        return '{}({!r}, {})'.format(
            self.__class__.__name__, self._dice_pool, self._select
        )


class BestOfDicePool(DicePoolCollection):
    """
    Take the best [select] rolls from a DicePool of [pool_size] * [input_die].
    BestOfDicePool(DicePool(Die(6), 4), 3) is the best 3 rolls from four six-sided dice.
    """

    def __init__(self, pool: DicePool, select: int):
        super().__init__(pool, select)

    def _generate_dict(self):
        ordered_combinations = self._dice_pool.rolls
        start_at = self._dice_pool.size - self._select
        stop_before = self._dice_pool.size
        return generate_events_dict(ordered_combinations, start_at, stop_before)


class WorstOfDicePool(DicePoolCollection):
    """
    Take the worst [select] rolls from a DicePool of [pool_size] * [input_die].
    WorstOfDicePool(DicePool(Die(6), 4), 3) is the worst 3 rolls from four six-sided dice.
    """

    def __init__(self, pool: DicePool, select: int):
        super().__init__(pool, select)

    def _generate_dict(self):
        ordered_combinations = self._dice_pool.rolls
        start_at = 0
        stop_before = self._select
        return generate_events_dict(ordered_combinations, start_at, stop_before)


class UpperMidOfDicePool(DicePoolCollection):
    """
    Take the middle [select] rolls from a DicePool of [pool_size] * [input_die].
    UpperMidOfDicePool(DicePool(Die(6), 5), 3) is the middle 3 rolls from five six-sided dice.

    If there is no perfect middle, take the higher of two choices. For five dice that roll
    (1, 1, 2, 3, 4), select=3 takes (1, 2, 3) and select=2 takes (2, 3).
    """

    def __init__(self, pool: DicePool, select: int):
        super().__init__(pool, select)

    def _generate_dict(self):
        ordered_combinations = self._dice_pool.rolls
        end_slice, extra = divmod(self._dice_pool.size - self._select, 2)
        start_at = end_slice + extra
        stop_before = self._dice_pool.size - end_slice
        return generate_events_dict(ordered_combinations, start_at, stop_before)


class LowerMidOfDicePool(DicePoolCollection):
    """
    Take the middle [select] rolls from a DicePool of [pool_size] * [input_die].
    LowerMidOfDicePool(DicePool(Die(6), 5), 3) is the middle 3 rolls from five six-sided dice.

    If there is no perfect middle, take the lower of two choices. For five dice that roll
    (1, 1, 2, 3, 4), select=3 takes (1, 2, 3) and select=2 takes (1, 2).
    """

    def __init__(self, pool: DicePool, select: int):
        super().__init__(pool, select)

    def _generate_dict(self):
        ordered_combinations = self._dice_pool.rolls
        end_slice, extra = divmod(self._dice_pool.size - self._select, 2)
        start_at = end_slice
        stop_before = self._dice_pool.size - (end_slice + extra)
        return generate_events_dict(ordered_combinations, start_at, stop_before)


def generate_events_dict(ordered_combinations, start_at, stop_before) -> Dict[int, int]:
    master_dict = {}
    for key, val in ordered_combinations.items():
        master_key = sum(key[start_at:stop_before])
        master_dict[master_key] = master_dict.get(master_key, 0) + val
    return master_dict



from typing import Dict, Tuple

from dicetables.eventsbases.protodie import ProtoDie
from dicetables.tools.orderedcombinations import ordered_combinations_of_events


class DicePool(object):
    def __init__(self, input_die: ProtoDie, pool_size: int):
        if pool_size < 1:
            raise ValueError("Minimum DicePool size is 1")
        self._input_die = input_die
        self._pool_size = pool_size
        self._rolls = ordered_combinations_of_events(input_die, pool_size)

    @property
    def die(self) -> ProtoDie:
        return self._input_die

    @property
    def size(self) -> int:
        return self._pool_size

    @property
    def rolls(self) -> Dict[Tuple[int, ...], int]:
        return self._rolls.copy()

    def __eq__(self, other):
        if not isinstance(other, DicePool):
            return False
        return (self.die, self.size) == (other.die, other.size)

    def __hash__(self):
        return hash((self.die, self.size))

    def __repr__(self):
        return "{}({!r}, {})".format(self.__class__.__name__, self.die, self.size)



"""
StaticDict is an immutable dictionary.
Getter is an object that remembers how to retrieve a parameter for a certain class or classes.
"""


class StaticDict(object):
    def __init__(self, dictionary):
        self._dic = dictionary.copy()

    def keys(self):
        return self._dic.keys()

    def values(self):
        return self._dic.values()

    def items(self):
        return self._dic.items()

    def copy(self):
        return StaticDict(self._dic)

    def delete(self, key):
        new = self._dic.copy()
        del new[key]
        return StaticDict(new)

    def get(self, key, default=None):
        return self._dic.get(key, default)

    def set(self, key, value):
        new = self._dic.copy()
        new[key] = value
        return StaticDict(new)


class Getter(object):
    def __init__(self, method_name, default_value, is_property=False):
        self._method_name = method_name
        self._is_property = is_property
        try:
            self._default_value = default_value.copy()
            self._copyable = True
        except AttributeError:
            self._default_value = default_value
            self._copyable = False

    def get_default(self):
        if self._copyable:
            return self._default_value.copy()
        else:
            return self._default_value

    def get_from(self, obj):
        if self._is_property:
            return obj.__getattribute__(self._method_name)
        else:
            return obj.__getattribute__(self._method_name)()

    def get_bool(self):
        return self._is_property

    def get_name(self):
        return self._method_name

    def __str__(self):
        is_property = {True: 'property', False: 'method'}
        return '{}: "{}", default: {!r}'.format(is_property[self._is_property], self._method_name, self._default_value)

    def __eq__(self, other):
        try:
            return ((self.get_default(), self.get_bool(), self.get_name()) ==
                    (other.get_default(), other.get_bool(), other.get_name()))
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other



import warnings


class EventsFactoryWarning(Warning):
    pass


class EventsFactoryWarningHandler(object):
    def __init__(self, factory):
        self._factory = factory

    def raise_warning(self, warning_code, *params):
        failed_class = params[0]
        msg_start = self._get_msg_start(failed_class, warning_code)
        msg_middle = self._get_error_code_body(params, warning_code)
        instructions = self._get_solution_message()

        warnings.warn(msg_start + msg_middle + instructions, EventsFactoryWarning, stacklevel=5)

    def _get_msg_start(self, failed_class, warning_code):
        msg_start = (
                '\nfactory: {}\n'.format(self._factory) +
                'Warning code: {}\n'.format(warning_code) +
                'Failed to find/add the following class to the EventsFactory - \n' +
                'class: {}\n'.format(failed_class)
        )
        return msg_start

    @staticmethod
    def _get_error_code_body(params, warning_code):
        msg_middle = ''
        if warning_code == 'CONSTRUCT':
            in_factory_class = params[1]
            msg_middle = (
                    '\nClass found in factory: {}\n'.format(in_factory_class) +
                    'attempted object construction using its signature. tried to return instance of original class.\n' +
                    'If that had failed, returned instance of the class found in EventsFactory.\n\n'
            )
        if warning_code == 'CHECK':
            msg_middle = (
                '\nWarning raised while performing check at instantiation\n\n'
            )
        return msg_middle

    def _get_solution_message(self):
        instructions = (
                'SOLUTION:\n' +
                '  class variable: factory_keys = (getter method/property names)\n'
                '  current factory keys are: {}\n'.format(self._factory.get_keys()[1]) +
                '  class variable: new_keys = [(info for each key not already in factory)]\n' +
                '  Each tuple in "new_keys" is (getter_name, default_value, "property"/"method")\n'
                'ex:\n' +
                '  NewClass(Something):\n' +
                '      factory_keys = ("dice_data", "get_dict", "get_thingy", "label")\n' +
                '      new_keys = [("get_thingy", 0, "method"),\n' +
                '                  ("label", "", "property")]\n\n' +
                '      def __init__(self, events_dict, dice_list, new_thingy, label):\n' +
                '          ....\n'
        )
        return instructions



"""
DiceTable and DetailedDiceTable compute combination of Die events
"""

from typing import TypeVar

from dicetables import DiceRecord
from dicetables.additiveevents import AdditiveEvents
from dicetables.eventsbases.protodie import ProtoDie
from dicetables.eventsinfo import EventsCalculations, EventsInformation
from dicetables.factory.eventsfactory import EventsFactory
from dicetables.tools.dictcombiner import DictCombiner

T = TypeVar('T', bound='DiceTable')


class DiceTable(AdditiveEvents):
    def __init__(self, events_dict: dict, dice_record: DiceRecord):
        self._record = dice_record
        super(DiceTable, self).__init__(events_dict)

    def dice_data(self) -> DiceRecord:
        return self._record

    def get_list(self):
        """

        :return: sorted copy of dice list: [(die, number of dice), ...]
        """
        return sorted(self._record.get_dict().items())

    def number_of_dice(self, query_die: ProtoDie) -> int:
        return self._record.get_number(query_die)

    def weights_info(self):
        """

        :return: str: complete info for all dice
        """
        output = ''
        for die, number in self.get_list():
            output += format_die_info(die, number)
        return output.rstrip('\n')

    def __str__(self):
        str_list = [die.multiply_str(number) for die, number in self.get_list()]
        return '\n'.join(str_list)

    def __repr__(self):
        class_name = self.__class__.__name__
        dice = str(self).replace('\n', ', ')
        return '<{} containing [{}]>'.format(class_name, dice)

    def add_die(self: T, die: ProtoDie, times: int = 1) -> T:
        """

        :param die: any subclass of ProtoDie: Die, ModDie, WeightedDie, ModWeightedDie, Modifier,
            StrongDie, Exploding, ExplodingOn
        :param times: int>= 0
        """
        dice_data = self._record.add_die(die, times)
        dictionary = DictCombiner(self.get_dict()).combine_by_fastest(die.get_dict(), times)
        return EventsFactory.from_dictionary_and_dice(self, dictionary, dice_data)

    def remove_die(self: T, die: ProtoDie, times=1) -> T:
        """

        :param die: any subclass of ProtoDie: Die, ModDie, WeightedDie, ModWeightedDie, Modifier,
            StrongDie, Exploding, ExplodingOn
        :param times: 0 <= int <= number of "die" in table
        """
        dice_data = self._record.remove_die(die, times)
        dictionary = DictCombiner(self.get_dict()).remove_by_tuple_list(die.get_dict(), times)
        return EventsFactory.from_dictionary_and_dice(self, dictionary, dice_data)

    def __eq__(self, other):
        return super(DiceTable, self).__eq__(other) and self.dice_data() == other.dice_data()


def format_die_info(die, number):
    weight_info = die.weight_info()
    adjusted_info = weight_info.replace(str(die), die.multiply_str(number)) + '\n\n'
    return adjusted_info


class DetailedDiceTable(DiceTable):
    def __init__(self, events_dict: dict, dice_record: DiceRecord, calc_includes_zeroes: bool = True):
        super(DetailedDiceTable, self).__init__(events_dict, dice_record)
        self._calc = EventsCalculations(self, calc_includes_zeroes)

    @property
    def info(self) -> EventsInformation:
        return self._calc.info

    @property
    def calc(self) -> EventsCalculations:
        return self._calc

    @property
    def calc_includes_zeroes(self) -> bool:
        return self._calc.include_zeroes

    def switch_boolean(self: T) -> T:
        """

        :return: a new DetailedDiceTable with the `calc_includes_zeroes` boolean switched
        """
        return EventsFactory.from_params(self, {'calc_includes_zeroes': not self.calc_includes_zeroes})

    def __eq__(self, other):
        return super(DetailedDiceTable, self).__eq__(other) and self.calc_includes_zeroes == other.calc_includes_zeroes



import ast
from typing import Dict, Iterable, Type, Union
from inspect import signature, _empty, Signature

from dicetables.dicepool_collection import (
    BestOfDicePool,
    WorstOfDicePool,
    UpperMidOfDicePool,
    LowerMidOfDicePool,
)
from dicetables.dicepool import DicePool
from dicetables.dieevents import (
    Die,
    ModDie,
    Modifier,
    ModWeightedDie,
    WeightedDie,
    StrongDie,
    Exploding,
    ExplodingOn,
)
from dicetables.eventsbases.protodie import ProtoDie
from dicetables.tools.limit_checker import (
    AbstractLimitChecker,
    NoOpLimitChecker,
    LimitChecker,
)


DieOrPool = Union[Type[ProtoDie], Type[DicePool]]


class ParseError(ValueError):
    pass


class Parser(object):
    def __init__(
        self,
        ignore_case: bool = False,
        checker: AbstractLimitChecker = NoOpLimitChecker(),
    ):
        """

        :param ignore_case: False: Can the parser ignore case on die names and kwargs.
        :param checker: `dicetables.tools.limit_checker.NoOpLimitChecker`: How limits will be enforced
            for parsing.  This defaults to a limit checker that does not check limits.
        """
        self.checker = checker
        self.ignore_case = ignore_case

        self._classes = {
            Die,
            ModDie,
            Modifier,
            ModWeightedDie,
            WeightedDie,
            StrongDie,
            Exploding,
            ExplodingOn,
            BestOfDicePool,
            WorstOfDicePool,
            UpperMidOfDicePool,
            LowerMidOfDicePool,
            DicePool,
        }
        self._param_types = {
            int: make_int,
            Dict[int, int]: make_int_dict,
            ProtoDie: self.make_die,
            Iterable[int]: make_int_tuple,
            DicePool: self.make_pool,
        }

    @classmethod
    def with_limits(
        cls,
        ignore_case: bool = False,
        max_size: int = 500,
        max_explosions: int = 10,
        max_dice: int = 6,
        max_dice_pools: int = 2,
    ) -> "Parser":
        """
        Creates a parser with a functioning limit checker from `dicetables.tools.limit_checker.LimitChecker`
        For explanation of how or why to change `max_dice_pool_combinations_per_dict_size`, and
        `max_dice_pool_calls`, see
        `Parser <http://dice-tables.readthedocs.io/en/latest/implementation_details/parser.html#limits-and-dicepool-objects>`_

        :param ignore_case: False: Can the parser ignore case on die names and kwargs.
        :param max_size: 500: The maximum allowed die size when :code:`parse_die_within_limits`
        :param max_explosions: 10: The maximum allowed (explosions + len(explodes_on)) when
            :code:`parse_die_within_limits`
        :param max_dice: 6: The maximum number of dice calls when :code:`parse`.
            Ex: :code:`StrongDie(Exploding(Die(5), 2), 3)` has 3 dice calls.
        :param max_dice_pools: 2: The maximum number of allowed dice_pool calls
        """
        checker = LimitChecker(
            max_dice_pools=max_dice_pools,
            max_dice=max_dice,
            max_explosions=max_explosions,
            max_size=max_size,
        )
        return cls(ignore_case=ignore_case, checker=checker)

    @property
    def param_types(self):
        return self._param_types.copy()

    @property
    def classes(self):
        return self._classes.copy()

    def parse_die(self, die_string):
        die_string = die_string.strip()
        ast_call_node = ast.parse(die_string).body[0].value

        self.checker.assert_numbers_of_calls_within_limits(
            self.walk_dice_calls(ast_call_node)
        )

        return self.make_die(ast_call_node)

    def make_die(self, call_node: ast.Call):
        die_class = self._get_call_class(call_node)
        bound_args = self._get_bound_args(call_node, die_class)

        self.checker.assert_explosions_within_limits(bound_args)
        self.checker.assert_die_size_within_limits(bound_args)

        return die_class(*bound_args.args, **bound_args.kwargs)

    def make_pool(self, call_node: ast.Call):
        pool_class = self._get_call_class(call_node)
        bound_args = self._get_bound_args(call_node, pool_class)

        self.checker.assert_dice_pool_within_limits(bound_args)

        return pool_class(*bound_args.args, **bound_args.kwargs)

    def walk_dice_calls(self, call_node: ast.AST) -> Iterable[DieOrPool]:
        return (
            self._get_call_class(node)
            for node in ast.walk(call_node)
            if isinstance(node, ast.Call)
        )

    def _get_call_class(self, call_node: ast.AST):
        class_name = call_node.func.id
        class_name = self._update_search_string(class_name)
        for die_class in self._classes:
            test_against = self._update_search_string(die_class.__name__)
            if class_name == test_against:
                return die_class
        raise ParseError("Die class: <{}> not recognized by parser.".format(class_name))

    def _update_search_string(self, search_str):
        if self.ignore_case:
            return search_str.lower()
        return search_str

    def _get_bound_args(self, call_node: ast.Call, call_class: DieOrPool):
        call_signature = signature(call_class)
        args = self._get_params(call_node, call_signature)
        kwargs = self._get_kwargs(call_node, call_signature)
        bound_args = call_signature.bind(*args, **kwargs)
        bound_args.apply_defaults()
        return bound_args

    def _get_params(self, call_node: ast.Call, call_signature: Signature):
        param_nodes = call_node.args
        params = []
        for node, param in zip(param_nodes, call_signature.parameters.values()):
            type_hint = param.annotation
            converter = self._param_types[type_hint]
            params.append(converter(node))
        return params

    def _raise_error_for_missing_type(self, call_signature: Signature):
        if any(
            param.annotation not in self._param_types
            for param in call_signature.parameters.values()
        ):
            raise ParseError(
                "The signature: {} has one or more un-recognized param types".format(
                    call_signature
                )
            )

    def _get_kwargs(self, call_node: ast.Call, call_signature: Signature):
        kwarg_nodes = call_node.keywords
        out = {}
        for kwarg_node in kwarg_nodes:
            kwarg_name, value = self._get_kwarg_value(call_signature, kwarg_node)
            out[kwarg_name] = value
        return out

    def _get_kwarg_value(self, die_signature, kwarg_node):
        kwarg_name_to_search_for = self._update_search_string(kwarg_node.arg)
        value_node = kwarg_node.value

        class_params = die_signature.parameters
        param_mapping = {
            self._update_search_string(key): value
            for key, value in class_params.items()
        }

        try:
            param = param_mapping[kwarg_name_to_search_for]
        except KeyError:
            raise ParseError(
                "The keyword: {} is not in the die signature: {}".format(
                    kwarg_node.arg, die_signature
                )
            )
        converter = self._param_types[param.annotation]

        return param.name, converter(value_node)

    def add_class(self, class_: object):
        """

        :param class_: the class you are adding
        """
        die_signature = signature(class_)
        self._raise_error_for_missing_annotation(die_signature)
        self._raise_error_for_missing_type(die_signature)
        self._classes.add(class_)

    @staticmethod
    def _raise_error_for_missing_annotation(die_signature: Signature):
        if any(
            param.annotation == _empty for param in die_signature.parameters.values()
        ):
            raise ParseError(
                "The signature: {} is missing type annotations".format(die_signature)
            )

    def add_param_type(self, param_type, creation_method):
        self._param_types[param_type] = creation_method


def make_int_dict(dict_node):
    keys = _make_int_list(dict_node.keys)
    values = _make_int_list(dict_node.values)
    return dict(zip(keys, values))


def make_int_tuple(tuple_node):
    return tuple(_make_int_list(tuple_node.elts))


def _make_int_list(num_node_list):
    return [make_int(num_node) for num_node in num_node_list]


def make_int(num_node):
    if isinstance(num_node, ast.UnaryOp) and isinstance(num_node.op, ast.USub):
        value = num_node.operand.n * -1
    else:
        value = None
        for key, val in ast.iter_fields(num_node):
            if key != "kind":  # pragma no branch
                value = val
    if not isinstance(value, int):
        raise ValueError("Expected an integer, but got: {!r}".format(value))
    return value



"""
tool for combining dictionaries.
IndexedValues is a list with a start index.  It is for simulating dictionaries of {int: int>0}
"""


def generate_indexed_values(sorted_tuple_list):
    """

    :param sorted_tuple_list: may not be empty.\n
        [(int, int>0), ...]
    """
    start_val, values = make_start_index_and_list(sorted_tuple_list)
    return IndexedValues(start_val, values)


def generate_indexed_values_from_dict(input_dict) -> 'IndexedValues':
    """

    :param input_dict: may not be empty.\n
        {int: int>0, ...}
    """
    return generate_indexed_values(sorted(input_dict.items()))


def make_start_index_and_list(sorted_tuple_list):
    start_val = sorted_tuple_list[0][0]
    end_val = sorted_tuple_list[-1][0]
    out_list = [0] * (end_val - start_val + 1)
    for index, value in sorted_tuple_list:
        list_index = index - start_val
        out_list[list_index] = value
    return start_val, out_list


class IndexedValues(object):
    def __init__(self, start_index, sorted_values):
        """

        :param start_index: int
        :param sorted_values: may not be empty\n
            [int>=0, ...] \n
            values[0] != 0\n
            values[-1] != 0
        """
        self._start_index = start_index
        self._values = sorted_values[:]

    @property
    def raw_values(self):
        return self._values[:]

    @property
    def start_index(self):
        return self._start_index

    @property
    def index_range(self):
        return self.start_index, len(self.raw_values) + self.start_index - 1

    def get_dict(self):
        return {index + self.start_index: value for index, value in enumerate(self.raw_values) if value}

    def get_value_at_key(self, key):
        index = key - self.start_index
        if index < 0 or index >= len(self.raw_values):
            return 0
        else:
            return self.raw_values[index]

    def combine_with_dictionary(self, no_zero_values_dict):
        """
        :param no_zero_values_dict: may not be empty\n
            {int: int>0, ...}
        """
        base_list = self.raw_values

        first_event = min(no_zero_values_dict.keys())
        last_event = max(no_zero_values_dict.keys())

        new_start_index = self.start_index + first_event
        new_size = len(base_list) + last_event - first_event

        container_for_lists_to_combine = []

        for event, occurrences in no_zero_values_dict.items():
            index_offset = event - first_event
            new_group_of_events = get_events_list(base_list, occurrences, new_size, index_offset)
            container_for_lists_to_combine.append(new_group_of_events)
        new_raw_values = list(map(add_many, *container_for_lists_to_combine))
        return IndexedValues(new_start_index, new_raw_values)


def add_many(*args):
    return sum(args)


def get_events_list(base_list, occurrences, new_size, index_offset):
    if occurrences != 1:
        adjusted_events = [value * occurrences for value in base_list]
    else:
        adjusted_events = base_list[:]
    return change_list_len_with_zeroes(adjusted_events, new_size, index_offset)


def change_list_len_with_zeroes(input_list, new_size, index_offset):
    left = [0] * index_offset
    right = [0] * (new_size - index_offset - len(input_list))
    return left + input_list + right



class EventsFactoryError(AssertionError):
    pass


class EventsFactoryErrorHandler(object):
    def __init__(self, factory):
        self._factory = factory
        self._format_kwargs = {}

    def raise_error(self, error_code, *params):
        self.create_format_kwargs(error_code, params)
        msg_header = self.create_header(error_code, params[0])
        msg_body = self.create_error_body(error_code)
        raise EventsFactoryError(msg_header + msg_body)

    def create_format_kwargs(self, error_code, params):
        """
        param_keys = {
            'CLASS OVERWRITE': ('events_class', 'class_keys', 'class_args_tuple'),
            'MISSING GETTER': ('events_class', 'current_getters', 'getter_key'),
            'GETTER OVERWRITE': ('getter_key', 'factory_getter', 'new_getter'),
            'SIGNATURES DIFFERENT': ('events_class', 'signature_in_factory'),
            'WTF': ('events_class', 'factory_classes')
        }
        """
        self._assign_parameters(error_code, params)
        self._assign_kwargs_from_factory_class()

    def _assign_parameters(self, error_code, params):
        param_values = {
            'CLASS OVERWRITE': ('events_class', 'class_args_tuple'),
            'MISSING GETTER': ('events_class', 'getter_key'),
            'GETTER OVERWRITE': ('getter_key', 'new_getter'),
            'SIGNATURES DIFFERENT': ('events_class',),
            'WTF': ('events_class',),
        }
        for index, key_name in enumerate(param_values[error_code]):
            self._format_kwargs[key_name] = params[index]

    def _assign_kwargs_from_factory_class(self):
        factory_classes, current_getters = self._factory.get_keys()
        kwargs_from_factory = {
            'factory_classes': factory_classes,
            'current_getters': current_getters
        }
        if 'events_class' in self._format_kwargs:
            kwargs_from_factory['class_keys'] = self._factory.get_class_params(self._format_kwargs['events_class'])
        if 'getter_key' in self._format_kwargs:
            kwargs_from_factory['factory_getter'] = self._factory.get_getter_string(self._format_kwargs['getter_key'])

        self._format_kwargs.update(kwargs_from_factory)

    def create_header(self, error_code, bad_param):
        msg_start = 'Error Code: {}\nFactory:    {}\nError At:   '.format(error_code, self._factory)
        if error_code == 'GETTER OVERWRITE':
            msg_end = 'Factory Getter Key: {!r}\n'.format(bad_param)
        else:
            msg_end = '{}\n'.format(bad_param)

        return msg_start + msg_end

    def create_error_body(self, error_code):
        explanation = get_explanation(error_code)
        format_str = get_format_string(error_code)
        param_details = format_str.format(**self._format_kwargs)
        return explanation + param_details


def get_explanation(error_code):
    explanation_body = '\nAttempted to {}.\n'
    explanation_variables = {
        'CLASS OVERWRITE': 'add class already in factory but used different factory keys',
        'MISSING GETTER': 'add class with a getter key not in the factory',
        'GETTER OVERWRITE': 'add getter key already in factory but used different parameters',
        'SIGNATURES DIFFERENT': 'construct a class already present in factory, but with a different signature',
        'WTF': 'construct a class unrelated, in any way, to any class in the factory'
    }
    explanation = explanation_body.format(explanation_variables[error_code])
    return explanation


def get_format_string(error_code):
    line_params = {
        'CLASS OVERWRITE': ('class', 'factory class keys', 'input class keys'),
        'MISSING GETTER': ('class', 'factory getters', 'missing getter'),
        'GETTER OVERWRITE': ('getter key', 'factory getter info', 'input getter info'),
        'SIGNATURES DIFFERENT': ('class', 'class signature', 'reset'),
        'WTF': ('factory classes', 'searched MRO')
    }
    return get_lines(line_params[error_code])


def get_lines(line_keys):
    lines = {
        'class': 'Class: {events_class}\n',
        'factory class keys': 'Current Factory Keys: {class_keys}\n',
        'input class keys': 'Keys Passed In:       {class_args_tuple}\n',
        'missing getter': 'Key Passed In:        {getter_key!r}\n',
        'getter key': 'Key: {getter_key!r}\n',
        'factory getter info': 'Factory Parameter:    {factory_getter}\n',
        'input getter info': 'Passed In Parameters: {new_getter}\n',
        'class signature': 'Signature In Factory: {class_keys}\n',
        'reset': 'To reset the factory to its base state, use EventsFactory.reset()\n',
        'factory getters': 'Current Factory Keys: {current_getters}\n',
        'factory classes': 'EventsFactory can currently construct the following classes:\n{factory_classes}\n',
        'searched MRO': ('EventsFactory searched the MRO of {events_class},\n' +
                         'and found no matches to the classes in the factory.\n')
    }
    out_str = ''
    for line in line_keys:
        out_str += lines[line]
    return out_str



from dicetables.additiveevents import AdditiveEvents
from dicetables.dicepool_collection import BestOfDicePool, WorstOfDicePool, UpperMidOfDicePool, LowerMidOfDicePool
from dicetables.dicepool import DicePool
from dicetables.dicerecord import DiceRecord
from dicetables.dicetable import DiceTable, DetailedDiceTable
from dicetables.dieevents import Die, ModDie, WeightedDie, ModWeightedDie, StrongDie, Modifier, ExplodingOn, Exploding
from dicetables.eventsbases.eventerrors import DiceRecordError, InvalidEventsError
from dicetables.eventsinfo import (EventsCalculations, EventsInformation,
                                   events_range, mean, stddev, percentage_points, percentage_axes, stats,
                                   full_table_string)
from dicetables.parser import Parser, ParseError
from dicetables.tools.limit_checker import LimitsError
from dicetables.roller import Roller



import random
from typing import Optional, List

from dicetables.tools.alias_table import AliasTable


class Roller(object):
    def __init__(self, events, random_generator: Optional[random.Random] = None):
        """

        :param events: any IntegerEvents
        :param random_generator: any instance of random.Random. defaults to python default random instance
        """
        self._alias_table = AliasTable(events.get_dict())
        self._random_generator = random_generator
        if not self._random_generator:
            self._random_generator = random.Random()

    @property
    def alias_table(self) -> AliasTable:
        """
        here is
        `a nice explanation of alias tables: <http://www.keithschwarz.com/darts-dice-coins/>`_

        :return: dicetables.tools.AliasTable
        """
        return self._alias_table

    @property
    def random_generator(self) -> random.Random:
        return self._random_generator

    def roll(self) -> int:
        length = self._random_generator.randrange(self._alias_table.length)
        height = self._random_generator.randrange(self._alias_table.height)
        return self._alias_table.get(length, height)

    def roll_many(self, times) -> List[int]:
        """

        :param times: int - how many times to roll
        :return: [int,..] - the value of each roll
        """
        return [self.roll() for _ in range(times)]



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods, line-too-long

from __future__ import absolute_import

import unittest
from decimal import Decimal

from dicetables.tools.numberforamtter import NumberFormatter, remove_extra_zero_from_single_digit_exponent


class TestNumberFormatter(unittest.TestCase):

    def assert_format_number(self, number, number_str):
        self.assertEqual(NumberFormatter().format(number), number_str)

    def test_NumberFormatter_init_defaults(self):
        test = NumberFormatter()
        self.assertEqual(test.min_fixed_pt_exp, -3)
        self.assertEqual(test.shown_digits, 4)
        self.assertEqual(test.max_comma_exp, 6)

    def test_NumberFormatter_shown_digits_setter(self):
        test = NumberFormatter()
        test.shown_digits = 2.5
        self.assertEqual(test.shown_digits, 2)
        test.shown_digits = -5
        self.assertEqual(test.shown_digits, 1)

    def test_NumberFormatter_max_comma_exp_setter(self):
        test = NumberFormatter()
        test.max_comma_exp = 2.5
        self.assertEqual(test.max_comma_exp, 2)
        test.max_comma_exp = -5
        self.assertEqual(test.max_comma_exp, -1)

    def test_NumberFormatter_min_fixed_pt_exp_setter(self):
        test = NumberFormatter()
        test.min_fixed_pt_exp = -5
        self.assertEqual(test.min_fixed_pt_exp, -5)
        test.min_fixed_pt_exp = 5
        self.assertEqual(test.min_fixed_pt_exp, 0)
        test.min_fixed_pt_exp = -5.2
        self.assertEqual(test.min_fixed_pt_exp, -5)

    def test_NumberFormatter_init_out_of_range(self):
        test = NumberFormatter(-1, -2, 1)
        self.assertEqual(test.shown_digits, 1)
        self.assertEqual(test.max_comma_exp, -1)
        self.assertEqual(test.min_fixed_pt_exp, 0)

    def test_NumberFormatter_init_in_range(self):
        test = NumberFormatter(5, 4, -5)
        self.assertEqual(test.min_fixed_pt_exp, -5)
        self.assertEqual(test.shown_digits, 5)
        self.assertEqual(test.max_comma_exp, 4)

    def test_NumberFormatter_get_exponent_between_pos_one_and_neg_one__big_small_pos_neg_no_rounding(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.get_exponent(1e-5), -5)
        self.assertEqual(formatter.get_exponent(-1e-5), -5)
        self.assertEqual(formatter.get_exponent(Decimal('1e-1000')), -1000)
        self.assertEqual(formatter.get_exponent(Decimal('-1e-1000')), -1000)

    def test_NumberFormatter_get_exponent_gt_one__float_pos_neg_no_rounding(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.get_exponent(1e+5), 5)
        self.assertEqual(formatter.get_exponent(-1e+5), 5)
        self.assertEqual(formatter.get_exponent(Decimal('1e+1000')), 1000)
        self.assertEqual(formatter.get_exponent(Decimal('-1e+1000')), 1000)

    def test_NumberFormatter_get_exponent_int__big_small_pos_neg(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.get_exponent(50), 1)
        self.assertEqual(formatter.get_exponent(-50), 1)
        self.assertEqual(formatter.get_exponent(5 * 10**1000), 1000)
        self.assertEqual(formatter.get_exponent(-5 * 10**1000), 1000)

    def test_NumberFormatter_get_exponent__rounding(self):
        four_digits = NumberFormatter()
        three_digits = NumberFormatter(shown_digits=3)
        self.assertEqual(four_digits.get_exponent(99.95), 1)
        self.assertEqual(three_digits.get_exponent(99.95), 2)

    def test_NumberFormatter_get_exponent__rounding_lt_one(self):
        four_digits = NumberFormatter()
        three_digits = NumberFormatter(shown_digits=3)
        self.assertEqual(four_digits.get_exponent(0.0099951), -3)
        self.assertEqual(three_digits.get_exponent(0.0099951), -2)

    def test_NumberFormatter_format_fixed_point_no_round(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.format_fixed_point(1.2e-5), '0.00001200')

    def test_NumberFormatter_format_fixed_point_round(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.format_fixed_point(9.9999e-5), '0.0001000')

    def test_NumberFormatter_format_fixed_point_neg(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.format_fixed_point(-9.9999e-5), '-0.0001000')

    def test_NumberFormatter_format_commaed_no_round(self):
        formatter = NumberFormatter()
        formatter.shown_digits = 10
        self.assertEqual(formatter.format_commaed(9999.99), '9,999.990000')

    def test_NumberFormatter_format_commaed_round(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.format_commaed(9999.99), '10,000')
        self.assertEqual(formatter.format_commaed(99.999), '100.0')

    def test_NumberFormatter_format_commaed_shows_number_as_int_if_too_long(self):
        formatter = NumberFormatter()
        formatter.shown_digits = 5
        self.assertEqual(formatter.format_commaed(123456.789), '123,457')

    def test_NumberFormatter_format_commaed_negative_number(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.format_commaed(-123456.789), '-123,457')

    def test_NumberFormatter_format_commaed_small_int(self):
        formatter = NumberFormatter()
        self.assertEqual(formatter.format_commaed(12), '12')

    def test_NumberFormatter_format_commaed_shows_whole_int(self):
        formatter = NumberFormatter()
        formatter.shown_digits = 5
        self.assertEqual(formatter.format_commaed(123456789), '123,456,789')

    def test_remove_extra_zero_Decimal_regression_test(self):
        """
        why does formatting a number stick in that damn extra zero???  Decimal has figured this out.
        it's not that hard. arrrrrggggghhhhhhhh!
        """
        float_format = '{:.2e}'.format(123)
        dec_format = '{:.2e}'.format(Decimal(123))
        dec_zero_exponent = '{:.2e}'.format(Decimal(1))
        self.assertEqual(float_format, '1.23e+02')
        self.assertEqual(dec_format, '1.23e+2')

        self.assertEqual(remove_extra_zero_from_single_digit_exponent(float_format), '1.23e+2')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent(dec_format), '1.23e+2')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent(dec_zero_exponent), '1.00e+0')

    def test_remove_extra_zero_from_exponent_removes(self):
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.23e-05'), '1.23e-5')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.23e+06'), '1.23e+6')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e+00'), '1.00e+0')

    def test_remove_extra_zero_from_exponent_does_not_remove(self):
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.23e-5'), '1.23e-5')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.23e+6'), '1.23e+6')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e+0'), '1.00e+0')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e+10'), '1.00e+10')

    def test_remove_extra_zero_from_exponent_safe_for_others(self):
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e+10'), '1.00e+10')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e+100'), '1.00e+100')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e-10'), '1.00e-10')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e-100'), '1.00e-100')
        self.assertEqual(remove_extra_zero_from_single_digit_exponent('1.00e+000'), '1.00e+000')

    def test_NumberFormatter_format_as_exponent_not_remove_extra_zero(self):
        formatter = NumberFormatter()
        test = 1.23456789 * 10**12
        self.assertEqual(formatter.format_exponent(test), '1.235e+12')

    def test_NumberFormatter_format_as_exponent_remove_extra_zero(self):
        formatter = NumberFormatter()
        test = 1.23456789 * 10**5
        self.assertEqual('{:.3e}'.format(test), '1.235e+05')
        self.assertEqual(formatter.format_exponent(test), '1.235e+5')

    def test_NumberFormatter_format_as_exponent_negative_number(self):
        formatter = NumberFormatter()
        test = -1.23456789 * 10**5
        self.assertEqual(formatter.format_exponent(test), '-1.235e+5')

    def test_NumberFormatter_format_as_exponent_lt_one(self):
        formatter = NumberFormatter()
        test = -1.23456789 * 10**(-5)
        self.assertEqual(formatter.format_exponent(test), '-1.235e-5')

    def test_NumberFormatter_white_box_test_OverflowError_calls_new_func_does_not_raise_error(self):
        formatter = NumberFormatter()
        test = 10**5000
        self.assertRaises(OverflowError, '{:e}'.format, test)
        self.assertEqual(formatter.format_exponent(test), '1.000e+5000')

    def test_NumberFormatter_format_huge_int_pos_rounds(self):
        formatter = NumberFormatter()
        test_1 = 123449 * 10**500
        test_2 = 123451 * 10**500
        self.assertEqual(formatter.format_exponent(test_1), '1.234e+505')
        self.assertEqual(formatter.format_exponent(test_2), '1.235e+505')

    def test_NumberFormatter_format_huge_int_neg_rounds(self):
        formatter = NumberFormatter()
        test_1 = -123449 * 10**500
        test_2 = -123451 * 10**500
        self.assertEqual(formatter.format_exponent(test_1), '-1.234e+505')
        self.assertEqual(formatter.format_exponent(test_2), '-1.235e+505')

    def test_NumberFormatter_format_huge_int_round_to_next_power(self):
        four_digits = NumberFormatter()
        eight_digits = NumberFormatter(shown_digits=8)
        test = 999951 * 10**500
        self.assertEqual(four_digits.format_exponent(test), '1.000e+506')
        self.assertEqual(eight_digits.format(test), '9.9995100e+505')

    def test_NumberFormatter_format_huge_int_round_to_next_power_negative_number(self):
        four_digits = NumberFormatter()
        eight_digits = NumberFormatter(shown_digits=8)
        test = -999951 * 10**500
        self.assertEqual(four_digits.format_exponent(test), '-1.000e+506')
        self.assertEqual(eight_digits.format(test), '-9.9995100e+505')

    def test_NumberFormatter_format_min_fixed_pt_exp_at_zero(self):
        test = NumberFormatter(min_fixed_pt_exp=0)
        self.assertEqual(test.format(0.1), '1.000e-1')

    def test_NumberFormatter_format_min_fixed_pt_exp_at_zero_max_comma_exp_at_zero(self):
        test = NumberFormatter(min_fixed_pt_exp=0, max_comma_exp=0)

        self.assertEqual(test.format(0.1), '1.000e-1')

        self.assertEqual(test.format(1.0), '1.000')
        self.assertEqual(test.format(1), '1')

        self.assertEqual(test.format(10.0), '1.000e+1')
        self.assertEqual(test.format(10), '1.000e+1')

    def test_NumberFormatter_format__min_fixed_pt_exp_at_zero_max_comma_exp_at_neg_one(self):
        test = NumberFormatter(max_comma_exp=-1, min_fixed_pt_exp=0)
        self.assertEqual(test.format(0.1), '1.000e-1')
        self.assertEqual(test.format(1.0), '1.000e+0')
        self.assertEqual(test.format(1), '1.000e+0')
        self.assertEqual(test.format(10.0), '1.000e+1')

    def test_NumberFormatter_max_comma_exp_ridiculous_case(self):
        test = NumberFormatter(max_comma_exp=1000)
        num = int('9' * 999)
        num_string = '999,' * 333
        num_string = num_string.rstrip(',')
        self.assertEqual(test.format(num), num_string)

    def test_NumberFormatter_format_zero(self):
        self.assert_format_number(0, '0')

    def test_NumberFormatter_format_edge_of_min_fixed_pt_exp(self):
        self.assert_format_number(0.000999949, '9.999e-4')
        self.assert_format_number(0.000999951, '0.001000')

    def test_NumberFormatter_format_edge_of_one_switches_algorithm(self):
        exponent_formatter = NumberFormatter(min_fixed_pt_exp=0)
        fixed_point_formatter = NumberFormatter(min_fixed_pt_exp=-5)
        self.assertEqual(exponent_formatter.format(.999949), '9.999e-1')
        self.assertEqual(exponent_formatter.format(.999951), '1.000')
        self.assertEqual(fixed_point_formatter.format(.999949), '0.9999')
        self.assertEqual(fixed_point_formatter.format(.999951), '1.000')

    def test_NumberFormatter_format_edge_of_max_comma_exp_int(self):
        self.assert_format_number(9999999, '9,999,999')

    def test_NumberFormatter_format_edge_of_max_comma_exp_float(self):
        self.assert_format_number(9999999.0, '1.000e+7')
        self.assert_format_number(9999499.0, '9,999,499')
        self.assert_format_number(9999501.0, '1.000e+7')

    def test_NumberFormatter_format_at_edge_of_max_commas_exp_with_large_show_digits(self):
        ten_digits = NumberFormatter(shown_digits=10)
        eight_digits = NumberFormatter(shown_digits=8)
        nine_digit_number = 9999999.99
        self.assertEqual(ten_digits.format(nine_digit_number), '9,999,999.990')
        self.assertEqual(eight_digits.format(nine_digit_number), '1.0000000e+7')

    def test_NumberFormatter_format_ints(self):
        self.assert_format_number(-3, '-3')
        self.assert_format_number(1000, '1,000')
        self.assert_format_number(-9999999, '-9,999,999')
        self.assert_format_number(10000000, '1.000e+7')
        self.assert_format_number(-99999999, '-1.000e+8')
        self.assert_format_number(999999999, '1.000e+9')
        self.assert_format_number(9999999999, '1.000e+10')
        self.assert_format_number(123451 * 10**1000, '1.235e+1005')

    def test_NumberFormatter_is_special_case_zero(self):
        self.assertTrue(NumberFormatter().is_special_case(0))
        self.assertTrue(NumberFormatter().is_special_case(0.0))

    def test_NumberFormatter_is_special_case_infinity(self):
        self.assertTrue(NumberFormatter().is_special_case(float('inf')))
        self.assertTrue(NumberFormatter().is_special_case(Decimal('inf')))

    def test_NumberFormatter_is_special_case_neg_infinity(self):
        self.assertTrue(NumberFormatter().is_special_case(float('-inf')))
        self.assertTrue(NumberFormatter().is_special_case(Decimal('-inf')))

    def test_NumberFormatter_is_special_case_false(self):
        self.assertFalse(NumberFormatter().is_special_case(1.23))
        self.assertFalse(NumberFormatter().is_special_case(-10 * 1111))
        self.assertFalse(NumberFormatter().is_special_case(Decimal('1e-1000')))
        self.assertFalse(NumberFormatter().is_special_case(Decimal('1e+1000')))
        self.assertFalse(NumberFormatter().is_special_case(Decimal('-1e+1000')))

    def test_NumberFormatter_get_special_case_zero(self):
        self.assertEqual(NumberFormatter().get_special_case(0), '0')
        self.assertEqual(NumberFormatter().get_special_case(0.0), '0')

    def test_NumberFormatter_get_special_case_infinity(self):
        self.assertEqual(NumberFormatter().get_special_case(float('inf')), 'Infinity')
        self.assertEqual(NumberFormatter().get_special_case(Decimal('inf')), 'Infinity')

    def test_NumberFormatter_get_special_case_neg_infinity(self):
        self.assertEqual(NumberFormatter().get_special_case(float('-inf')), '-Infinity')
        self.assertEqual(NumberFormatter().get_special_case(Decimal('-inf')), '-Infinity')

    def test_NumberFormatter_format_special_cases(self):
        self.assertEqual(NumberFormatter().format(0), '0')
        self.assertEqual(NumberFormatter().format(float('inf')), 'Infinity')
        self.assertEqual(NumberFormatter().format(float('-inf')), '-Infinity')

    def test_NumberFormatter_format_fixed_point_special_cases(self):
        self.assertEqual(NumberFormatter().format_fixed_point(0), '0')
        self.assertEqual(NumberFormatter().format_fixed_point(float('inf')), 'Infinity')
        self.assertEqual(NumberFormatter().format_fixed_point(float('-inf')), '-Infinity')

    def test_NumberFormatter_format_commaed_special_cases(self):
        self.assertEqual(NumberFormatter().format_commaed(0), '0')
        self.assertEqual(NumberFormatter().format_commaed(float('inf')), 'Infinity')
        self.assertEqual(NumberFormatter().format_commaed(float('-inf')), '-Infinity')

    def test_NumberFormatter_format_exponent_special_cases(self):
        self.assertEqual(NumberFormatter().format_exponent(0), '0')
        self.assertEqual(NumberFormatter().format_exponent(float('inf')), 'Infinity')
        self.assertEqual(NumberFormatter().format_exponent(float('-inf')), '-Infinity')


if __name__ == '__main__':
    unittest.main()



"""
AdditiveEvents is the general case for DiceTable - an IntegerEvents that can combine with other IntegerEvents.
"""
from typing import TypeVar, Type, Dict

from dicetables.eventsbases.integerevents import IntegerEvents
from dicetables.factory.eventsfactory import EventsFactory
from dicetables.tools.dictcombiner import DictCombiner


def scrub_zeroes(dictionary):
    return {key: val for key, val in dictionary.items() if val}


T = TypeVar('T', bound='AdditiveEvents')


class AdditiveEvents(IntegerEvents):

    def __init__(self, events_dict):
        """

        :param events_dict: {event: occurrences}\n
            event=int. occurrences=int >=0
            total occurrences > 0
        """
        self._table = scrub_zeroes(events_dict)
        super(AdditiveEvents, self).__init__()
        EventsFactory.check(self.__class__)

    @classmethod
    def new(cls: Type[T]) -> T:
        return EventsFactory.new(cls)

    def get_dict(self) -> Dict[int, int]:
        return self._table.copy()

    def __str__(self):
        min_event = min(self._table.keys())
        max_event = max(self._table.keys())
        return 'table from {} to {}'.format(min_event, max_event)

    def combine(self: T, events: IntegerEvents, times: int = 1) -> T:
        dictionary = DictCombiner(self.get_dict()).combine_by_fastest(events.get_dict(), times)
        return EventsFactory.from_dictionary(self, dictionary)

    def combine_by_flattened_list(self: T, events: IntegerEvents, times: int = 1) -> T:
        """

        :WARNING - UNSAFE METHOD: len(flattened_list) = total occurrences of events.
            if this list is too big, it will raise MemoryError or OverflowError
        """
        dictionary = DictCombiner(self.get_dict()).combine_by_flattened_list(events.get_dict(), times)
        return EventsFactory.from_dictionary(self, dictionary)

    def combine_by_dictionary(self: T, events: IntegerEvents, times: int = 1) -> T:
        dictionary = DictCombiner(self.get_dict()).combine_by_dictionary(events.get_dict(), times)
        return EventsFactory.from_dictionary(self, dictionary)

    def combine_by_indexed_values(self: T, events: IntegerEvents, times: int = 1) -> T:
        dictionary = DictCombiner(self.get_dict()).combine_by_indexed_values(events.get_dict(), times)
        return EventsFactory.from_dictionary(self, dictionary)

    def remove(self: T, events: IntegerEvents, times: int = 1) -> T:
        """

        :WARNING - UNSAFE METHOD: There is no record of what you added to an AdditiveEvents.
            If you remove what you haven't added, no error will be raised, but you will have bugs.
        """
        dictionary = DictCombiner(self.get_dict()).remove_by_tuple_list(events.get_dict(), times)
        return EventsFactory.from_dictionary(self, dictionary)



from setuptools import setup, find_packages


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='dicetables',
      version='4.0.1',
      description='get all combinations for any set of dice',
      long_description=readme(),
      keywords='dice, die, statistics, table, probability, combinations',
      url='http://github.com/eric-s-s/dice-tables',
      author='Eric Shaw',
      author_email='shaweric01@gmail.com',
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',
          "Operating System :: OS Independent",
          'Intended Audience :: Developers',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Games/Entertainment :: Role-Playing',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
      ],
      packages=find_packages(exclude=['tests', 'time_trials', 'docs']),
      install_requires=[],
      include_package_data=True,
      zip_safe=False)



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import

import unittest

from dicetables.additiveevents import AdditiveEvents
from dicetables.dicerecord import DiceRecord
from dicetables.dicetable import DetailedDiceTable
from dicetables.factory.factorytools import StaticDict, Getter


class TestFactoryTools(unittest.TestCase):
    def test_StaticDict_init_empty(self):
        empty = StaticDict({})
        self.assertEqual(empty.items(), {}.items())

    def test_StaticDict_init_non_empty_items(self):
        new = StaticDict({1: 2, 3: 4})
        self.assertEqual(new.items(), {1: 2, 3: 4}.items())

    def test_StaticDict_init_is_not_original(self):
        input_dict = {1: 2, 3: 4}
        new = StaticDict(input_dict)
        input_dict[1] = 'a'
        self.assertEqual(new.items(), {1: 2, 3: 4}.items())

    def test_StaticDict_keys_empty(self):
        empty = StaticDict({})
        self.assertEqual(empty.keys(), {}.keys())

    def test_StaticDict_keys(self):
        new = StaticDict({1: 2, 3: 4})
        self.assertEqual(tuple(new.keys()), (1, 3))

    def test_StaticDict_values_empty(self):
        empty = StaticDict({})
        self.assertEqual(empty.keys(), {}.keys())

    def test_StaticDict_values(self):
        new = StaticDict({1: 2, 3: 4})
        self.assertEqual(tuple(new.values()), (2, 4))

    def test_StaticDict_copy_empty(self):
        new = StaticDict({})
        cpy = new.copy()
        self.assertIsNot(cpy, new)
        self.assertEqual(cpy.items(), {}.items())

    def test_StaticDict_copy(self):
        new = StaticDict({1: 2, 3: 4})
        cpy = new.copy()
        self.assertEqual(cpy.items(), new.items())

    def test_StaticDict_copy_returns_correct_type(self):
        new = StaticDict({})
        self.assertIs(type(new.copy()), StaticDict)

    def test_StaticDict_get_not_there_default_not_set(self):
        new = StaticDict({1: 2, 3: 4})
        answer = new.get(2)
        self.assertIsNone(answer)

    def test_StaticDict_get_not_there_default_set(self):
        new = StaticDict({1: 2, 3: 4})
        answer = new.get(2, 'a')
        self.assertEqual(answer, 'a')

    def test_StaticDict_get_returns_correct_value(self):
        new = StaticDict({1: 2, 3: 4})
        self.assertEqual(new.get(1), 2)

    def test_StaticDict_set_new_key(self):
        new = StaticDict({1: 2, 3: 4})
        new = new.set(4, 5)
        self.assertEqual(tuple(new.items()), ((1, 2), (3, 4), (4, 5)))

    def test_StaticDict_set_present_key(self):
        new = StaticDict({1: 2, 3: 4})
        new = new.set(3, 5)
        self.assertEqual(tuple(new.items()), ((1, 2), (3, 5)))

    def test_StaticDict_set_returns_correct_type(self):
        new = StaticDict({})
        self.assertIs(type(new.set(1, 2)), StaticDict)

    def test_StaticDict_delete_key_not_present_raises_key_error(self):
        new = StaticDict({1: 2, 3: 4})
        self.assertRaises(KeyError, new.delete, 5)

    def test_StaticDict_delete(self):
        new = StaticDict({1: 2, 3: 4})
        new = new.delete(3)
        self.assertEqual(tuple(new.items()), ((1, 2),))

    def test_StaticDict_delete_returns_correct_type(self):
        new = StaticDict({1: 2})
        self.assertIs(type(new.delete(1)), StaticDict)

    def test_Getter_init_copies_if_possible(self):
        new = {}
        getter = Getter('hi', new)
        new[1] = 2
        self.assertEqual(getter.get_default(), {})

    def test_Getter_get_default(self):
        self.assertEqual(Getter('get_dict', {0: 1}).get_default(), {0: 1})

    def test_Getter_get_default_does_not_copy_if_no_copy_method(self):
        self.assertEqual(Getter('a', 5).get_default(), 5)

    def test_Getter_get_default_copies_if_possible(self):
        dic = {0: 1}
        getter = Getter('get_dict', dic)
        default = getter.get_default()
        default[1] = 5
        dic[2] = 7
        self.assertEqual(getter.get_default(), {0: 1})

    def test_Getter_get_from__object_has_getter_method(self):
        events = AdditiveEvents({1: 1})
        self.assertEqual(Getter('get_dict', {0: 1}).get_from(events), {1: 1})

    def test_Getter_get_from__object_has_getter_property(self):
        events = DetailedDiceTable({1: 1}, DiceRecord.new(), calc_includes_zeroes=False)
        getter = Getter('calc_includes_zeroes', True, is_property=True)
        self.assertEqual(getter.get_from(events), False)

    def test_Getter_get_bool_true(self):
        self.assertTrue(Getter('a', 1, is_property=True).get_bool())

    def test_Getter_get_bool_false(self):
        self.assertFalse(Getter('a', 1, is_property=False).get_bool())

    def test_Getter_get_name(self):
        self.assertEqual(Getter('a', 1).get_name(), 'a')

    def test_Getter_str_method_property(self):
        expected_string = 'property: "calc_includes_zeroes", default: True'
        self.assertEqual(str(Getter('calc_includes_zeroes', True, is_property=True)), expected_string)

    def test_Getter_str_method_method(self):
        expected_string = 'method: "calc_includes_zeroes", default: True'
        self.assertEqual(str(Getter('calc_includes_zeroes', True)), expected_string)

    def test_Getter_eq_true(self):
        getter1 = Getter('get_num', 0)
        getter2 = Getter('get_num', 0)
        self.assertTrue(getter1 == getter2)

    def test_Getter_eq_false_by_default(self):
        getter1 = Getter('get_num', 0)
        getter2 = Getter('get_num', '0')
        self.assertFalse(getter1 == getter2)

    def test_Getter_eq_false_by_incomparable_defaults(self):
        class Bob(object):
            number = 5

            def __eq__(self, other):
                return self.number == other.number

        class Frank(object):
            letter = 'a'

            def __eq__(self, other):
                return self.letter == other.letter

        self.assertRaises(AttributeError, Bob().__eq__, Frank())
        self.assertFalse(Getter('a', Bob()) == Getter('a', Frank()))

    def test_Getter_eq_false_by_defaults_object_has_no_eq_method_two_instances(self):
        class Bob(object):
            number = 5

        bob1 = Bob()
        bob2 = Bob()
        self.assertFalse(Getter('a', bob1) == Getter('a', bob2))

    def test_Getter_eq_false_by_property_bool(self):
        getter1 = Getter('get_num', 0, is_property=True)
        getter2 = Getter('get_num', 0, is_property=False)
        self.assertFalse(getter1 == getter2)

    def test_Getter_eq_false_by_name(self):
        getter1 = Getter('get_number', 0)
        getter2 = Getter('get_num', 0)
        self.assertFalse(getter1 == getter2)

    def test_Getter_ne_false(self):
        getter1 = Getter('get_num', 0)
        getter2 = Getter('get_num', 0)
        self.assertFalse(getter1 != getter2)

    def test_Getter_ne_true(self):
        getter1 = Getter('get_number', 0)
        getter2 = Getter('get_num', 0)
        self.assertTrue(getter1 != getter2)


if __name__ == '__main__':
    unittest.main()



"""
The abstract class for any die represented by a set of events
"""
from dicetables.eventsbases.integerevents import IntegerEvents


class ProtoDie(IntegerEvents):
    """
    This is the basis for all the dice classes.

    all Die objects need:

    - get_size() - returns: int > 0
    - get_weight() - returns: int >= 0
    - weight_info() - returns: str
    - multiply_str() - returns: str
    - get_dict() - returns: {int: int > 0}
    - __repr__
    - __str__
    """

    def __init__(self):
        super(ProtoDie, self).__init__()

    def get_size(self) -> int:
        raise NotImplementedError

    def get_weight(self) -> int:
        raise NotImplementedError

    def get_dict(self):
        super(ProtoDie, self).get_dict()

    def weight_info(self) -> str:
        """return detailed info of how the die is weighted"""
        raise NotImplementedError

    def multiply_str(self, number: int) -> str:
        """return a string that is the die string multiplied by a number. i.e.,
        D6+1 times 3 is '3D6+3' """
        raise NotImplementedError

    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        raise NotImplementedError

    def __hash__(self):
        return hash('hash of {!r}, {}, {}, {}'.format(self,
                                                      self.get_size(),
                                                      self.get_weight(),
                                                      self.get_dict()))

    def __lt__(self, other):
        return (
                (self.get_size(), self.get_weight(), sorted(self.get_dict().items()), repr(self)) <
                (other.get_size(), other.get_weight(), sorted(other.get_dict().items()), repr(other))
        )

    def __eq__(self, other):
        return (super(ProtoDie, self).__eq__(other) and
                (self.get_size(), self.get_weight(), repr(self)) ==
                (other.get_size(), other.get_weight(), repr(other))
                )

    def __le__(self, other):
        return self < other or self == other

    def __gt__(self, other):
        return not self <= other

    def __ge__(self, other):
        return self == other or self > other



from abc import ABC, abstractmethod
from enum import Enum
from inspect import BoundArguments
from typing import Iterable, Type, Optional, Any, Union

from dicetables.dicepool import DicePool
from dicetables.eventsbases.protodie import ProtoDie
from dicetables.tools.orderedcombinations import (
    count_unique_combination_keys,
    largest_permitted_pool_size,
)


DieOrPool = Union[Type[ProtoDie], Type[DicePool]]


class LimitsError(ValueError):
    def __init__(self, *args):
        super(LimitsError, self).__init__(*args)


class ArgumentType(Enum):
    SIZE = 1
    EXPLOSIONS = 2
    EXPLODES_ON = 3
    INPUT_DIE = 4
    POOL_SIZE = 5


def get_bound_args(arg_type: ArgumentType, bound_args: BoundArguments) -> Optional[Any]:
    arg_names = {
        ArgumentType.SIZE: ("die_size", "dictionary_input"),
        ArgumentType.EXPLOSIONS: ("explosions",),
        ArgumentType.EXPLODES_ON: ("explodes_on",),
        ArgumentType.INPUT_DIE: ("input_die",),
        ArgumentType.POOL_SIZE: ("pool_size",),
    }
    possible_args = arg_names[arg_type]
    for arg_name, arg_value in bound_args.arguments.items():
        if arg_name in possible_args:
            return arg_value
    return None


class AbstractLimitChecker(ABC):
    @abstractmethod
    def assert_numbers_of_calls_within_limits(
        self, die_classes: Iterable[DieOrPool]
    ) -> None:
        """
        asserts that the number of `dicetables.ProtoDie` calls and the number of
        `dicetables.dicepool.DicePool` calls are within limits.

        :raises LimitsError:
        """
        raise NotImplementedError

    @abstractmethod
    def assert_die_size_within_limits(self, bound_args: BoundArguments) -> None:
        """
        Checks the bound arguments for the size of the die and asserts they are
        within limits.

        This typically uses `dicetables.tools.limit_checker.get_bound_args` to determine what
        is a `dicetables.tools.limit_checker.ArgumentType.SIZE` argument.

        :raises LimitsError:
        """
        raise NotImplementedError

    @abstractmethod
    def assert_explosions_within_limits(self, bound_args: BoundArguments) -> None:
        """
        Asserts that the number of explosions on an `dicetables.Exploding` or an `dicetables.ExplodingOn`
        has a number of explosions withing limits.

        This typically uses `dicetables.tools.limit_checker.get_bound_args` to determine what
        is a `dicetables.tools.limit_checker.ArgumentType.EXPLOSIONS` and
        a `dicetables.tools.limit_checker.ArgumentType.EXPLODES_ON` argument.

        :raises LimitsError:
        """
        raise NotImplementedError

    @abstractmethod
    def assert_dice_pool_within_limits(self, bound_args: BoundArguments) -> None:
        """
        asserts that the size of a dice pool is within limits.  Dice pools can be very expensive
        to calculate. see:

        `DicePools <http://dice-tables.readthedocs.io/en/latest/the_dice.html#dice-pools>`_

        This typically uses `dicetables.tools.limit_checker.get_bound_args` to determine what
        is a `dicetables.tools.limit_checker.ArgumentType.POOL_SIZE` argument.

        :raises LimitsError:
        """
        raise NotImplementedError


class NoOpLimitChecker(AbstractLimitChecker):
    def assert_numbers_of_calls_within_limits(
        self, die_classes: Iterable[DieOrPool]
    ) -> None:
        pass

    def assert_die_size_within_limits(self, bound_args: BoundArguments) -> None:
        pass

    def assert_explosions_within_limits(self, bound_args: BoundArguments) -> None:
        pass

    def assert_dice_pool_within_limits(self, bound_args: BoundArguments) -> None:
        pass


class LimitChecker(AbstractLimitChecker):
    def __init__(
        self,
        max_size: int = 500,
        max_explosions: int = 10,
        max_dice: int = 5,
        max_dice_pools: int = 2,
    ):
        self.max_size = max_size
        self.max_explosions = max_explosions
        self.max_dice_calls = max_dice
        self.max_dice_pool_calls = max_dice_pools
        self.max_dice_pool_combinations_per_dict_size = {
            2: 600,
            3: 8700,
            4: 30000,
            5: 55000,
            6: 70000,
            7: 100000,
            12: 200000,
            30: 250000,
        }

    def assert_numbers_of_calls_within_limits(
        self, call_classes: Iterable[DieOrPool]
    ) -> None:
        class_list = list(call_classes)
        die_classes = len([el for el in class_list if issubclass(el, ProtoDie)])
        dice_pools = class_list.count(DicePool)
        if (
            die_classes > self.max_dice_calls
            or dice_pools > self.max_dice_pool_calls
        ):
            msg = (
                "Limits exceeded. Max dice calls: {}. ".format(self.max_dice_calls) +
                "Max dice pool calls: {}. ".format(self.max_dice_pool_calls) +
                "Calls requested: {}".format(class_list)
            )
            raise LimitsError(msg)

    def assert_die_size_within_limits(self, bound_args: BoundArguments) -> None:
        sized_value = get_bound_args(ArgumentType.SIZE, bound_args)
        if isinstance(sized_value, dict):
            size = max(sized_value.keys())
        else:
            size = sized_value

        if size and size > self.max_size:
            raise LimitsError(
                "A die of size: {} is greater than the allowed ".format(size) +
                "max: {}".format(self.max_size)
            )

    def assert_explosions_within_limits(self, bound_args: BoundArguments) -> None:
        explosions = get_bound_args(ArgumentType.EXPLOSIONS, bound_args)
        explodes_on = get_bound_args(ArgumentType.EXPLODES_ON, bound_args)
        if explodes_on:
            explosions += len(explodes_on)

        if explosions and explosions > self.max_explosions:
            raise LimitsError(
                (
                    "Explosions: {} + ".format(explosions) +
                    "len({}) is greater than allowed ".format(explodes_on) +
                    "max: {}".format(self.max_explosions)
                )
            )

    def assert_dice_pool_within_limits(self, bound_args: BoundArguments) -> None:
        input_die = get_bound_args(ArgumentType.INPUT_DIE, bound_args)
        pool_size = get_bound_args(ArgumentType.POOL_SIZE, bound_args)
        if input_die is None or pool_size is None:
            return

        dict_size = len(input_die.get_dict())
        pool_limit = 0
        for key, limit in sorted(self.max_dice_pool_combinations_per_dict_size.items()):
            if key > dict_size:
                break
            pool_limit = limit

        score = count_unique_combination_keys(input_die, pool_size)
        if score > pool_limit:
            max_pool_size = largest_permitted_pool_size(input_die, pool_limit)
            msg = "{!r} has a get_dict() of size: {}\n".format(input_die, dict_size)
            explanation = "For this die, the largest permitted pool_size is {}".format(
                max_pool_size
            )
            raise LimitsError(msg + explanation)



"""
fast consistent and programmable object to convert all number types to strings.

- gives consistent rounding and output between int, float and Decimal.
- can do exponential notation for ints over 10**310 significantly faster than Decimal
- can be set to automatically change from fixed to commaed to exponential notation at specified cutoffs.
"""

from math import log10


class NumberFormatter(object):
    def __init__(self, shown_digits=4, max_comma_exp=6, min_fixed_pt_exp=-3):
        """

        :param shown_digits: int >= 1
        :param max_comma_exp: int >= -1
        :param min_fixed_pt_exp: int <= 0
        """
        self._shown_digits = None
        self._max_comma_exp = None
        self._min_fixed_pt_exp = None
        self._special_cases = {0: '0', float('inf'): 'Infinity', float('-inf'): '-Infinity'}

        self.shown_digits = shown_digits
        self.max_comma_exp = max_comma_exp
        self.min_fixed_pt_exp = min_fixed_pt_exp

    @property
    def shown_digits(self):
        return self._shown_digits

    @shown_digits.setter
    def shown_digits(self, sig_figs):
        self._shown_digits = int(max(1, sig_figs))

    @property
    def max_comma_exp(self):
        return self._max_comma_exp

    @max_comma_exp.setter
    def max_comma_exp(self, value):
        self._max_comma_exp = int(max(-1, value))

    @property
    def min_fixed_pt_exp(self):
        return self._min_fixed_pt_exp

    @min_fixed_pt_exp.setter
    def min_fixed_pt_exp(self, value):
        self._min_fixed_pt_exp = min(0, int(value))

    def format(self, number):
        if self.is_special_case(number):
            return self.get_special_case(number)
        exponent = self.get_exponent(number)
        if 0 > exponent >= self.min_fixed_pt_exp:
            return self._format_number_and_exponent_to_fixed_point(number, exponent)
        elif 0 <= exponent <= self.max_comma_exp:
            return self._format_number_and_exponent_to_commas(number, exponent)
        else:
            return self._format_number_and_exponent_to_exponent(number, exponent)

    def format_fixed_point(self, number):
        """

        :param number: -1 < number < 1
        """
        if self.is_special_case(number):
            return self.get_special_case(number)
        exponent = self.get_exponent(number)
        return self._format_number_and_exponent_to_fixed_point(number, exponent)

    def format_commaed(self, number):
        """

        :param number: number >= 1 or number <= -1
        """
        if self.is_special_case(number):
            return self.get_special_case(number)
        exponent = self.get_exponent(number)
        return self._format_number_and_exponent_to_commas(number, exponent)

    def format_exponent(self, number):
        if self.is_special_case(number):
            return self.get_special_case(number)
        exponent = self.get_exponent(number)
        return self._format_number_and_exponent_to_exponent(number, exponent)

    def is_special_case(self, number):
        return number in self._special_cases

    def get_special_case(self, number):
        return self._special_cases[number]

    def get_exponent(self, number):
        if isinstance(number, int):
            return int(log10(abs(number)))
        return int('{:.{}e}'.format(number, self.shown_digits - 1).split('e')[1])

    def _format_number_and_exponent_to_fixed_point(self, number, exponent):
        return '{:.{}f}'.format(number, self.shown_digits - 1 - exponent)

    def _format_number_and_exponent_to_commas(self, number, exponent):
        if isinstance(number, int):
            return '{:,}'.format(number)
        else:
            return '{:,.{}f}'.format(number, max(0, self.shown_digits - 1 - exponent))

    def _format_number_and_exponent_to_exponent(self, number, exponent):
        try:
            answer = '{:.{}e}'.format(number, self.shown_digits - 1)
            if -10 < exponent < 10:
                return remove_extra_zero_from_single_digit_exponent(answer)
            return answer
        except OverflowError:
            return self._format_huge_int_and_exponent_to_exponent(number, exponent)

    def _format_huge_int_and_exponent_to_exponent(self, number, exponent):
        extra_digits = 10
        mantissa = number // 10 ** (exponent - self.shown_digits - extra_digits)
        mantissa /= 10. ** (self.shown_digits + extra_digits)
        mantissa = round(mantissa, self.shown_digits - 1)
        if mantissa == 10.0 or mantissa == -10.0:
            mantissa /= 10.0
            exponent += 1
        return '{:.{}f}e+{}'.format(mantissa, self.shown_digits - 1, exponent)


def remove_extra_zero_from_single_digit_exponent(answer):
    if answer[-2] == '0' and answer[-3] in ('-', '+'):
        return answer[:-2] + answer[-1:]
    return answer
